'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = undefined;

var _promise = require('babel-runtime/core-js/promise');

var _promise2 = _interopRequireDefault(_promise);

var _asyncToGenerator2 = require('babel-runtime/helpers/asyncToGenerator');

var _asyncToGenerator3 = _interopRequireDefault(_asyncToGenerator2);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

var _set = require('babel-runtime/core-js/set');

var _set2 = _interopRequireDefault(_set);

var _neoBlockchainNetwork = require('neo-blockchain-network');

var _neoBlockchainCore = require('neo-blockchain-core');

var _neoBlockchainNodeCore = require('neo-blockchain-node-core');

var _bloomFilter = require('bloom-filter');

var _bloomFilter2 = _interopRequireDefault(_bloomFilter);

var _bloem = require('bloem');

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _net = require('net');

var _net2 = _interopRequireDefault(_net);

var _Command = require('./Command');

var _payload = require('./payload');

var _Message = require('./Message');

var _Message2 = _interopRequireDefault(_Message);

var _errors = require('./errors');

require('./PeerData');

var _package = require('../package.json');

var _package2 = _interopRequireDefault(_package);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const createPeerBloomFilter = ({
  filter,
  k,
  tweak
}) => new _bloomFilter2.default({
  vData: Buffer.from(filter),
  nHashFuncs: k,
  nTweak: tweak
});


const createScalingBloomFilter = () => new _bloem.ScalingBloem(0.05, {
  initial_capacity: 100000,
  scaling: 4
});

const MEM_POOL_SIZE = 30000;
const GET_ADDR_PEER_COUNT = 200;
const GET_BLOCKS_COUNT = 500;
// Assume that we get 500 back, but if not, at least request every 10 seconds
const GET_BLOCKS_BUFFER = GET_BLOCKS_COUNT / 3;
const GET_BLOCKS_TIME_MS = 10000;
const GET_BLOCKS_THROTTLE_MS = 500;
const GET_BLOCKS_CLOSE_COUNT = 3;
const LOCAL_HOST_ADDRESSES = new _set2.default(['0.0.0.0', 'localhost', '127.0.0.1']);

let Node = function () {
  function Node(options) {
    var _this = this;

    (0, _classCallCheck3.default)(this, Node);

    this._negotiate = (() => {
      var _ref = (0, _asyncToGenerator3.default)(function* (peer) {
        _this._sendMessage(peer, _this._createMessage({
          command: _Command.COMMAND.VERSION,
          payload: new _payload.VersionPayload({
            protocolVersion: 0,
            services: _payload.SERVICES.NODE_NETWORK,
            timestamp: Math.round(Date.now() / 1000),
            port: _this._externalPort,
            nonce: _this._nonce,
            userAgent: _this._userAgent,
            startHeight: _this._blockchain.currentBlockIndex,
            relay: true
          })
        }));

        const message = yield peer.receiveMessage(30000);
        let versionPayload;
        if (message.value.command === _Command.COMMAND.VERSION) {
          versionPayload = message.value.payload;
        } else {
          throw new _errors.NegotiationError(message);
        }

        if (!_this._checkVersion(versionPayload)) {
          throw new _errors.NegotiationError(message);
        }

        let address;
        const host = (0, _neoBlockchainNodeCore.getEndpointConfig)(peer.endpoint).host;
        if (_net2.default.isIPv4(host)) {
          address = new _payload.NetworkAddress({
            host,
            port: versionPayload.port,
            timestamp: versionPayload.timestamp,
            services: versionPayload.services
          });
        }

        _this._sendMessage(peer, _this._createMessage({ command: _Command.COMMAND.VERACK }));

        const nextMessage = yield peer.receiveMessage(30000);
        if (nextMessage.value.command !== _Command.COMMAND.VERACK) {
          throw new _errors.NegotiationError(nextMessage);
        }

        return {
          data: {
            nonce: versionPayload.nonce,
            startHeight: versionPayload.startHeight,
            bloomFilter: null,
            address
          },
          relay: versionPayload.relay
        };
      });

      return function (_x) {
        return _ref.apply(this, arguments);
      };
    })();

    this._onEvent = event => {
      if (event.event === 'PEER_CONNECT_SUCCESS') {
        const { connectedPeer } = event.extra;
        if (this._bestPeer == null || this._bestPeer.data.startHeight < connectedPeer.data.startHeight) {
          this._bestPeer = connectedPeer;
          this._requestBlocks();
        }
      } else if (event.event === 'PEER_CLOSED') {
        if (this._bestPeer === event.extra.peer) {
          this._bestPeer = this._findBestPeer();
          this._requestBlocks();
        }
      }

      if (event.meta == null) {
        this._blockchain.logger({
          event: event.event,
          level: 'debug',
          message: event.message,
          context: this._loggingContext
        });
      } else {
        this._blockchain.logger({
          event: event.event,
          level: event.meta.type === 'error' ? undefined : 'debug',
          message: event.message,
          meta: event.meta,
          context: this._loggingContext
        });
      }
    };

    this._requestBlocks = _lodash2.default.throttle(_lodash2.default.debounce(() => {
      const peer = this._bestPeer;
      const block = this._blockchain.currentBlock;
      if (peer != null && block.index < peer.data.startHeight) {
        if (this._getBlocksRequestsCount > GET_BLOCKS_CLOSE_COUNT) {
          peer.close();
          this._getBlocksRequestsCount = 0;
        } else if (this._shouldRequestBlocks()) {
          if (this._getBlocksRequestsIndex === block.index) {
            this._getBlocksRequestsCount += 1;
          } else {
            this._getBlocksRequestsCount = 1;
            this._getBlocksRequestsIndex = block.index;
          }
          this._getBlocksRequestTime = Date.now();
          this._sendMessage(peer, this._createMessage({
            command: _Command.COMMAND.GET_BLOCKS,
            payload: new _payload.GetBlocksPayload({
              hashStart: [block.hash]
            })
          }));
        }

        this._requestBlocks();
      }
    }, GET_BLOCKS_THROTTLE_MS), GET_BLOCKS_THROTTLE_MS);

    this._onRequestEndpoints = () => {
      this._relay(this._createMessage({ command: _Command.COMMAND.GET_ADDR }));
    };

    this._blockchain = options.blockchain;
    this._network = new _neoBlockchainNetwork.Network({
      seeds: options.seeds,
      listenTCP: options.listenTCP,
      externalEndpoints: options.externalEndpoints,
      connectPeersDelayMS: options.connectPeersDelayMS,
      maxConnectedPeers: options.maxConnectedPeers,
      socketTimeoutMS: options.socketTimeoutMS,
      negotiate: this._negotiate,
      createMessageTransform: () => new _Message.MessageTransform(this._blockchain.deserializeWireContext),
      onMessageReceived: this._onMessageReceived.bind(this),
      onRequestEndpoints: this._onRequestEndpoints.bind(this),
      onEvent: this._onEvent
    });

    this._started = false;
    this._stopped = false;
    this._loggingContext = {
      type: 'node',
      identifier: this._blockchain.identifier
    };

    this._externalPort = (options.listenTCP || {}).port || 0;
    this._nonce = Math.floor(Math.random() * _neoBlockchainCore.utils.UINT_MAX_NUMBER);
    this._userAgent = `NEO:neo-blockchain-js:${_package2.default.version}`;

    this.memPool = {};
    this._knownBlockHashes = createScalingBloomFilter();
    this._tempKnownBlockHashes = new _set2.default();
    this._knownTransactionHashes = createScalingBloomFilter();
    this._tempKnownTransactionHashes = new _set2.default();
    this._knownHeaderHashes = createScalingBloomFilter();
    this._tempKnownHeaderHashes = new _set2.default();
    this._getBlocksRequestsIndex = null;
    this._getBlocksRequestTime = null;
    this._getBlocksRequestsCount = 1;
  }

  (0, _createClass3.default)(Node, [{
    key: 'start',
    value: function start() {
      if (this._started) {
        return;
      }
      this._started = true;
      this._stopped = false;

      this._blockchain.logger({
        event: 'NODE_START',
        message: 'Node starting...',
        context: this._loggingContext
      });
      try {
        this._network.start();
      } catch (error) {
        this._started = false;
        throw error;
      }
    }
  }, {
    key: 'stop',
    value: function stop() {
      if (this._stopped) {
        return;
      }
      this._stopped = true;

      this._blockchain.logger({
        event: 'NODE_STOP',
        message: 'Node stopping...',
        context: this._loggingContext
      });
      try {
        this._network.stop();
        this._started = false;
      } catch (error) {
        this._stopped = false;
        throw error;
      }
    }
  }, {
    key: 'relayTransaction',
    value: (() => {
      var _ref2 = (0, _asyncToGenerator3.default)(function* (transaction) {
        if (transaction.type === _neoBlockchainCore.TRANSACTION_TYPE.MINER || this.memPool[transaction.hashHex] != null || this._tempKnownTransactionHashes.has(transaction.hashHex)) {
          return;
        }

        if (!this._knownTransactionHashes.has(transaction.hash)) {
          this._blockchain.logger({
            event: 'RELAY_TRANSACTION_START',
            level: 'debug',
            message: `Relaying transaction: ${_neoBlockchainCore.common.uInt256ToString(transaction.hash)}`,
            meta: { type: 'extra', transaction },
            context: this._loggingContext
          });

          this._tempKnownTransactionHashes.add(transaction.hashHex);

          try {
            const foundTransaction = yield this._blockchain.transaction.tryGet({
              hash: transaction.hash
            });
            if (foundTransaction == null) {
              yield this._blockchain.verifyTransaction({
                transaction,
                memPool: _neoBlockchainCore.utils.values(this.memPool)
              });
              this.memPool[transaction.hashHex] = transaction;
              this._knownTransactionHashes.add(transaction.hash);
              this._relayTransaction(transaction);
              this._blockchain.logger({
                event: 'RELAY_TRANSACTION_SUCCESS',
                level: 'debug',
                message: `Relayed transaction: ` + `${_neoBlockchainCore.common.uInt256ToString(transaction.hash)}`,
                meta: { type: 'extra', transaction },
                context: this._loggingContext
              });
              yield this._trimMemPool();
            }
          } catch (error) {
            this._blockchain.logger({
              event: 'RELAY_TRANSACTION_ERROR',
              message: `Relaying transaction ` + `${_neoBlockchainCore.common.uInt256ToString(transaction.hash)} encountered ` + `error: ${error.message}`,
              meta: { type: 'error', error, extra: { transaction } },
              context: this._loggingContext
            });

            throw error;
            // eslint-disable-next-line
          } finally {
            this._tempKnownTransactionHashes.delete(transaction.hashHex);
          }
        }
      });

      function relayTransaction(_x2) {
        return _ref2.apply(this, arguments);
      }

      return relayTransaction;
    })()
  }, {
    key: '_relay',
    value: function _relay(message) {
      this._blockchain.logger({
        event: 'RELAY_MESSAGE',
        level: 'debug',
        message: `Relaying message: ${message.value.command}`,
        meta: { type: 'extra', message },
        context: this._loggingContext
      });
      this._network.relay(message.serializeWire());
    }
  }, {
    key: '_relayTransaction',
    value: function _relayTransaction(transaction) {
      const message = this._createMessage({
        command: _Command.COMMAND.INV,
        payload: new _payload.InvPayload({
          type: _payload.INVENTORY_TYPE.TRANSACTION,
          hashes: [transaction.hash]
        })
      });
      const messagePayload = message.serializeWire();
      for (const peer of this._network.connectedPeers) {
        if (peer.relay && this._testFilter(peer.data.bloomFilter, transaction)) {
          peer.write(messagePayload);
        }
      }
    }
  }, {
    key: '_sendMessage',
    value: function _sendMessage(peer, message) {
      this._blockchain.logger({
        event: 'SEND_MESSAGE',
        level: 'debug',
        message: `Sending message "${message.value.command}" to peer at ${peer.endpoint}`,
        meta: { type: 'extra', peer, message },
        context: this._loggingContext
      });
      peer.write(message.serializeWire());
    }
  }, {
    key: '_findBestPeer',
    value: function _findBestPeer() {
      return _lodash2.default.maxBy(this._network.connectedPeers, peer => peer.data.startHeight);
    }
  }, {
    key: '_shouldRequestBlocks',
    value: function _shouldRequestBlocks() {
      const block = this._blockchain.currentBlock;
      const getBlocksRequestTime = this._getBlocksRequestTime;
      return this._getBlocksRequestsIndex == null || block.index - this._getBlocksRequestsIndex > GET_BLOCKS_BUFFER || getBlocksRequestTime == null || Date.now() - getBlocksRequestTime > GET_BLOCKS_TIME_MS;
    }
  }, {
    key: '_checkVersion',
    value: function _checkVersion(version) {
      if (version.nonce === this._nonce) {
        return false;
      }

      if (this._network.connectedPeers.some(peer => version.nonce === peer.data.nonce)) {
        return false;
      }

      return true;
    }
  }, {
    key: '_onMessageReceived',
    value: function _onMessageReceived(peer, message) {
      this._blockchain.logger({
        event: 'MESSAGE_RECEIVED',
        level: 'debug',
        message: `Received "${message.value.command}" from ${peer.endpoint}.`,
        meta: { type: 'extra', peer, message },
        context: this._loggingContext
      });

      const onError = error => {
        this._blockchain.logger({
          event: 'MESSAGE_RECEIVED_ERROR',
          message: `Error while processing "${message.value.command}" from ` + `${peer.endpoint}`,
          meta: { type: 'error', error, extra: { peer, message } },
          context: this._loggingContext
        });
      };

      try {
        switch (message.value.command) {
          case _Command.COMMAND.ADDR:
            this._onAddrMessageReceived(message.value.payload);
            break;
          case _Command.COMMAND.BLOCK:
            this._onBlockMessageReceived(message.value.payload).catch(onError);
            break;
          case _Command.COMMAND.CONSENSUS:
            this._onConsensusMessageReceived(message.value.payload);
            break;
          case _Command.COMMAND.FILTER_ADD:
            this._onFilterAddMessageReceived(peer, message.value.payload);
            break;
          case _Command.COMMAND.FILTER_CLEAR:
            this._onFilterClearMessageReceived(peer);
            break;
          case _Command.COMMAND.FILTER_LOAD:
            this._onFilterLoadMessageReceived(peer, message.value.payload);
            break;
          case _Command.COMMAND.GET_ADDR:
            this._onGetAddrMessageReceived(peer);
            break;
          case _Command.COMMAND.GET_BLOCKS:
            this._onGetBlocksMessageReceived(peer, message.value.payload).catch(onError);
            break;
          case _Command.COMMAND.GET_DATA:
            this._onGetDataMessageReceived(peer, message.value.payload).catch(onError);
            break;
          case _Command.COMMAND.GET_HEADERS:
            this._onGetHeadersMessageReceived(peer, message.value.payload).catch(onError);
            break;
          case _Command.COMMAND.HEADERS:
            this._onHeadersMessageReceived(peer, message.value.payload).catch(onError);
            break;
          case _Command.COMMAND.INV:
            this._onInvMessageReceived(peer, message.value.payload);
            break;
          case _Command.COMMAND.MEMPOOL:
            this._onMemPoolMessageReceived(peer);
            break;
          case _Command.COMMAND.TX:
            this._onTransactionReceived(message.value.payload).catch(onError);
            break;
          case _Command.COMMAND.VERACK:
            this._onVerackMessageReceived(peer);
            break;
          case _Command.COMMAND.VERSION:
            this._onVersionMessageReceived(peer);
            break;
          case _Command.COMMAND.ALERT:
            break;
          case _Command.COMMAND.MERKLE_BLOCK:
            break;
          case _Command.COMMAND.NOT_FOUND:
            break;
          case _Command.COMMAND.PING:
            break;
          case _Command.COMMAND.PONG:
            break;
          case _Command.COMMAND.REJECT:
            break;
          default:
            // eslint-disable-next-line
            message.value.command;
            break;
        }
      } catch (error) {
        onError(error);
      }
    }
  }, {
    key: '_onAddrMessageReceived',
    value: function _onAddrMessageReceived(addr) {
      addr.addresses.filter(address => !LOCAL_HOST_ADDRESSES.has(address.host)).map(address => (0, _neoBlockchainNodeCore.createEndpoint)({
        type: 'tcp',
        host: address.host,
        port: address.port
      })).forEach(endpoint => this._network.addEndpoint(endpoint));
    }
  }, {
    key: '_onBlockMessageReceived',
    value: (() => {
      var _ref3 = (0, _asyncToGenerator3.default)(function* (block) {
        if (this._blockchain.currentBlockIndex >= block.index || this._tempKnownBlockHashes.has(block.hashHex)) {
          return;
        }

        if (!this._knownBlockHashes.has(block.hash)) {
          this._tempKnownBlockHashes.add(block.hashHex);

          try {
            const foundBlock = yield this._blockchain.block.tryGet({
              hashOrIndex: block.hash
            });
            if (foundBlock == null) {
              yield this._blockchain.persistBlock({ block });

              const peer = this._bestPeer;
              if (peer != null && block.index > peer.data.startHeight) {
                this._relay(this._createMessage({
                  command: _Command.COMMAND.INV,
                  payload: new _payload.InvPayload({
                    type: _payload.INVENTORY_TYPE.BLOCK,
                    hashes: [block.hash]
                  })
                }));
              }
            }

            this._knownBlockHashes.add(block.hash);
            this._knownHeaderHashes.add(block.hash);
            for (const transaction of block.transactions) {
              delete this.memPool[transaction.hashHex];
              this._knownTransactionHashes.add(transaction.hash);
            }
          } finally {
            this._tempKnownBlockHashes.delete(block.hashHex);
          }
        }
      });

      function _onBlockMessageReceived(_x3) {
        return _ref3.apply(this, arguments);
      }

      return _onBlockMessageReceived;
    })()

    // TODO: Implement
    // eslint-disable-next-line

  }, {
    key: '_onConsensusMessageReceived',
    value: function _onConsensusMessageReceived(consensus) {}

    // eslint-disable-next-line

  }, {
    key: '_onFilterAddMessageReceived',
    value: function _onFilterAddMessageReceived(peer, filterAdd) {
      if (peer.data.bloomFilter != null) {
        peer.data.bloomFilter.insert(filterAdd.data);
      }
    }

    // eslint-disable-next-line

  }, {
    key: '_onFilterClearMessageReceived',
    value: function _onFilterClearMessageReceived(peer) {
      // eslint-disable-next-line
      peer.data.bloomFilter = null;
    }

    // eslint-disable-next-line

  }, {
    key: '_onFilterLoadMessageReceived',
    value: function _onFilterLoadMessageReceived(peer, filterLoad) {
      // eslint-disable-next-line
      peer.data.bloomFilter = createPeerBloomFilter(filterLoad);
    }
  }, {
    key: '_onGetAddrMessageReceived',
    value: function _onGetAddrMessageReceived(peer) {
      const addresses = _lodash2.default.take(_lodash2.default.shuffle(this._network.connectedPeers.map(connectedPeer => connectedPeer.data.address).filter(Boolean)), GET_ADDR_PEER_COUNT);
      if (addresses.length > 0) {
        this._sendMessage(peer, this._createMessage({
          command: _Command.COMMAND.ADDR,
          payload: new _payload.AddrPayload({ addresses })
        }));
      }
    }
  }, {
    key: '_onGetBlocksMessageReceived',
    value: (() => {
      var _ref4 = (0, _asyncToGenerator3.default)(function* (peer, getBlocks) {
        const headers = yield this._getHeaders(getBlocks, this._blockchain.currentBlockIndex);
        this._sendMessage(peer, this._createMessage({
          command: _Command.COMMAND.INV,
          payload: new _payload.InvPayload({
            type: _payload.INVENTORY_TYPE.BLOCK,
            hashes: headers.map(function (header) {
              return header.hash;
            })
          })
        }));
      });

      function _onGetBlocksMessageReceived(_x4, _x5) {
        return _ref4.apply(this, arguments);
      }

      return _onGetBlocksMessageReceived;
    })()
  }, {
    key: '_onGetDataMessageReceived',
    value: (() => {
      var _ref5 = (0, _asyncToGenerator3.default)(function* (peer, getData) {
        var _this2 = this;

        switch (getData.type) {
          case 0x01:
            // Transaction
            yield _promise2.default.all(getData.hashes.map((() => {
              var _ref6 = (0, _asyncToGenerator3.default)(function* (hash) {
                let transaction = _this2.memPool[_neoBlockchainCore.common.uInt256ToHex(hash)];
                if (transaction == null) {
                  transaction = yield _this2._blockchain.transaction.tryGet({ hash });
                }

                if (transaction != null) {
                  _this2._sendMessage(peer, _this2._createMessage({
                    command: _Command.COMMAND.TX,
                    payload: transaction
                  }));
                }
              });

              return function (_x8) {
                return _ref6.apply(this, arguments);
              };
            })()));
            break;
          case 0x02:
            // Block
            yield _promise2.default.all(getData.hashes.map((() => {
              var _ref7 = (0, _asyncToGenerator3.default)(function* (hash) {
                const block = yield _this2._blockchain.block.tryGet({
                  hashOrIndex: hash
                });
                if (block != null) {
                  if (peer.data.bloomFilter == null) {
                    _this2._sendMessage(peer, _this2._createMessage({
                      command: _Command.COMMAND.BLOCK,
                      payload: block
                    }));
                  } else {
                    _this2._sendMessage(peer, _this2._createMessage({
                      command: _Command.COMMAND.MERKLE_BLOCK,
                      payload: _this2._createMerkleBlockPayload({
                        block,
                        flags: block.transactions.map(function (transaction) {
                          return _this2._testFilter(peer.data.bloomFilter, transaction);
                        })
                      })
                    }));
                  }
                }
              });

              return function (_x9) {
                return _ref7.apply(this, arguments);
              };
            })()));
            break;
          case 0xe0:
            // Consensus
            // TODO: Implement
            break;
          default:
            // eslint-disable-next-line
            getData.type;
            break;
        }
      });

      function _onGetDataMessageReceived(_x6, _x7) {
        return _ref5.apply(this, arguments);
      }

      return _onGetDataMessageReceived;
    })()
  }, {
    key: '_onGetHeadersMessageReceived',
    value: (() => {
      var _ref8 = (0, _asyncToGenerator3.default)(function* (peer, getBlocks) {
        const headers = yield this._getHeaders(getBlocks, this._blockchain.currentHeader.index);
        this._sendMessage(peer, this._createMessage({
          command: _Command.COMMAND.HEADERS,
          payload: new _payload.HeadersPayload({ headers })
        }));
      });

      function _onGetHeadersMessageReceived(_x10, _x11) {
        return _ref8.apply(this, arguments);
      }

      return _onGetHeadersMessageReceived;
    })()
  }, {
    key: '_onHeadersMessageReceived',
    value: (() => {
      var _ref9 = (0, _asyncToGenerator3.default)(function* (peer, headersPayload) {
        var _this3 = this;

        const headers = headersPayload.headers.filter(function (header) {
          return !_this3._knownHeaderHashes.has(header.hash) && !_this3._tempKnownHeaderHashes.has(header.hashHex);
        });
        if (headers.length > 0) {
          for (const header of headers) {
            this._tempKnownHeaderHashes.add(header.hashHex);
          }
          try {
            yield this._blockchain.persistHeaders(headers);
            for (const header of headers) {
              this._knownHeaderHashes.add(header.hash);
            }
          } finally {
            for (const header of headers) {
              this._tempKnownHeaderHashes.delete(header.hashHex);
            }
          }
        }

        if (this._blockchain.currentHeader.index < peer.data.startHeight) {
          this._sendMessage(peer, this._createMessage({
            command: _Command.COMMAND.GET_HEADERS,
            payload: new _payload.GetBlocksPayload({
              hashStart: [this._blockchain.currentHeader.hash]
            })
          }));
        }
      });

      function _onHeadersMessageReceived(_x12, _x13) {
        return _ref9.apply(this, arguments);
      }

      return _onHeadersMessageReceived;
    })()
  }, {
    key: '_onInvMessageReceived',
    value: function _onInvMessageReceived(peer, inv) {
      let hashes;
      switch (inv.type) {
        case 0x01:
          // Transaction
          hashes = inv.hashes.filter(hash => !this._knownTransactionHashes.has(hash) && !this._tempKnownTransactionHashes.has(_neoBlockchainCore.common.uInt256ToHex(hash)));
          break;
        case 0x02:
          // Block
          hashes = inv.hashes.filter(hash => !this._knownBlockHashes.has(hash) && !this._tempKnownBlockHashes.has(_neoBlockchainCore.common.uInt256ToHex(hash)));
          break;
        case 0xe0:
          // Consensus
          // TODO: Implement
          hashes = [];
          break;
        default:
          // eslint-disable-next-line
          inv.type;
          hashes = [];
          break;
      }

      if (hashes.length > 0) {
        this._sendMessage(peer, this._createMessage({
          command: _Command.COMMAND.GET_DATA,
          payload: new _payload.InvPayload({ type: inv.type, hashes })
        }));
      }
    }
  }, {
    key: '_onMemPoolMessageReceived',
    value: function _onMemPoolMessageReceived(peer) {
      this._sendMessage(peer, this._createMessage({
        command: _Command.COMMAND.INV,
        payload: new _payload.InvPayload({
          type: _payload.INVENTORY_TYPE.TRANSACTION,
          hashes: _neoBlockchainCore.utils.values(this.memPool).map(transaction => transaction.hash)
        })
      }));
    }
  }, {
    key: '_onTransactionReceived',
    value: (() => {
      var _ref10 = (0, _asyncToGenerator3.default)(function* (transaction) {
        yield this.relayTransaction(transaction);
      });

      function _onTransactionReceived(_x14) {
        return _ref10.apply(this, arguments);
      }

      return _onTransactionReceived;
    })()

    // eslint-disable-next-line

  }, {
    key: '_onVerackMessageReceived',
    value: function _onVerackMessageReceived(peer) {
      peer.close();
    }

    // eslint-disable-next-line

  }, {
    key: '_onVersionMessageReceived',
    value: function _onVersionMessageReceived(peer) {
      peer.close();
    }
  }, {
    key: '_getHeaders',
    value: (() => {
      var _ref11 = (0, _asyncToGenerator3.default)(function* (getBlocks, maxHeight) {
        var _this4 = this;

        let hashStopIndexPromise = _promise2.default.resolve(maxHeight);
        if (getBlocks.hashStop !== 0) {
          hashStopIndexPromise = this._blockchain.header.tryGet({ hashOrIndex: getBlocks.hashStop }).then(function (hashStopHeader) {
            return hashStopHeader == null ? maxHeight : Math.min(hashStopHeader.index, maxHeight);
          });
        }
        const [hashStartHeaders, hashEnd] = yield _promise2.default.all([_promise2.default.all(getBlocks.hashStart.map(function (hash) {
          return _this4._blockchain.header.tryGet({ hashOrIndex: hash });
        })), hashStopIndexPromise]);
        const hashStartHeader = _lodash2.default.head(_lodash2.default.orderBy(hashStartHeaders.filter(Boolean), [function (header) {
          return header.index;
        }]));
        if (hashStartHeader == null) {
          return [];
        }
        const hashStart = hashStartHeader.index + 1;
        if (hashStart > maxHeight) {
          return [];
        }

        const headers = yield _promise2.default.all(_lodash2.default.range(hashStart, Math.min(hashStart + GET_BLOCKS_COUNT, hashEnd)).map(function (index) {
          return _this4._blockchain.header.get({ hashOrIndex: index });
        }));

        return headers;
      });

      function _getHeaders(_x15, _x16) {
        return _ref11.apply(this, arguments);
      }

      return _getHeaders;
    })()
  }, {
    key: '_trimMemPool',
    value: (() => {
      var _ref12 = (0, _asyncToGenerator3.default)(function* () {
        var _this5 = this;

        const memPool = _neoBlockchainCore.utils.values(this.memPool);
        if (memPool.length > MEM_POOL_SIZE) {
          const transactionAndFee = yield _promise2.default.all(memPool.map((() => {
            var _ref13 = (0, _asyncToGenerator3.default)(function* (transaction) {
              const networkFee = yield transaction.getNetworkFee({
                getOutput: _this5._blockchain.output.get,
                governingToken: _this5._blockchain.settings.governingToken,
                utilityToken: _this5._blockchain.settings.utilityToken,
                fees: _this5._blockchain.settings.fees
              });
              return [transaction, networkFee];
            });

            return function (_x17) {
              return _ref13.apply(this, arguments);
            };
          })()));
          const hashesToRemove = _lodash2.default.take(_lodash2.default.sortBy(transactionAndFee,
          // TODO: Might be a bug since we're converting to number
          function ([transaction, networkFee]) {
            return networkFee.divn(transaction.size).toNumber();
          })
          // eslint-disable-next-line
          ).map(function ([transaction, _]) {
            return transaction.hashHex;
          });
          for (const hash of hashesToRemove) {
            delete this.memPool[hash];
          }
        }
      });

      function _trimMemPool() {
        return _ref12.apply(this, arguments);
      }

      return _trimMemPool;
    })()
  }, {
    key: '_testFilter',
    value: function _testFilter(bloomFilterIn, transaction) {
      const bloomFilter = bloomFilterIn;
      if (bloomFilter == null) {
        return true;
      }
      return bloomFilter.contains(transaction.hash) || transaction.outputs.some(output => bloomFilter.contains(output.address)) || transaction.inputs.some(input => bloomFilter.contains(input.serializeWire())) || transaction.scripts.some(script => bloomFilter.contains(_neoBlockchainCore.crypto.toScriptHash(script.verification))) || transaction.type === _neoBlockchainCore.TRANSACTION_TYPE.REGISTER && transaction instanceof _neoBlockchainCore.RegisterTransaction && bloomFilter.contains(transaction.asset.admin);
    }
  }, {
    key: '_createMerkleBlockPayload',
    value: function _createMerkleBlockPayload({
      block,
      flags
    }) {
      const tree = new _neoBlockchainCore.MerkleTree(block.transactions.map(transaction => transaction.hash)).trim(flags);

      const buffer = Buffer.allocUnsafe(Math.floor((flags.length + 7) / 8));
      for (let i = 0; i < flags.length; i += 1) {
        if (flags[i]) {
          // eslint-disable-next-line
          buffer[Math.floor(i / 8)] |= 1 << i % 8;
        }
      }

      return new _payload.MerkleBlockPayload({
        version: block.version,
        previousHash: block.previousHash,
        merkleRoot: block.merkleRoot,
        timestamp: block.timestamp,
        index: block.index,
        consensusData: block.consensusData,
        nextConsensus: block.nextConsensus,
        script: block.script,
        transactionCount: block.transactions.length,
        hashes: tree.toHashArray(),
        flags: buffer
      });
    }
  }, {
    key: '_createMessage',
    value: function _createMessage(value) {
      return new _Message2.default({
        magic: this._blockchain.settings.messageMagic,
        value
      });
    }
  }, {
    key: 'connectedPeersCount',
    get: function () {
      return this._network.connectedPeers.length;
    }
  }]);
  return Node;
}();

exports.default = Node;